"""
Common utilities for unit testing.
"""

import json
import os
import textwrap
from pathlib import Path
from typing import Optional, Tuple, Union, List

import yaml
from pydantic import Field, BaseModel, ConfigDict

from kugl.impl.config import Settings
from kugl.impl.engine import Engine, Query
from kugl.impl.registry import Registry
from kugl.util import to_utc, UNIT_TEST_TIMEBASE


def kubectl_response(kind: str, output: Union[str, dict]):
    """
    Put a mock response for 'kubectl get {kind} ...' into the mock responses folder,
    to be found by an invocation of ./kubectl in a test.
    :param kind: e.g. "pods", "nodes, "jobs" etc
    :param output: A dict (will be JSON-serialized) or a string (will be trimmed)
    """
    if isinstance(output, dict):
        output = json.dumps(output)
    else:
        output = str(output).strip()
    folder = Path(os.getenv("KUGL_MOCKDIR"))
    folder.mkdir(exist_ok=True)
    folder.joinpath(kind).write_text(output)


class Taint(BaseModel):
    """Helper class for creating taints in test nodes"""
    key: str
    effect: str
    value: Optional[str] = None


class CGM(BaseModel):
    """Helper class for creating CPU/GPU/Memory resources in test containers"""
    model_config = ConfigDict(populate_by_name=True)
    cpu: Union[int, str, None] = None
    mem: Union[int, str, None] = Field(None, alias="memory")
    gpu: Union[int, str, None] = Field(None, alias="nvidia.com/gpu")


class Container(BaseModel):
    """Helper class for creating containers in test pods"""
    name: str = "main"
    command: List[str] = Field(default_factory = lambda: ["echo", "hello"])
    requests: Optional[CGM] = CGM(cpu=1, mem="10M")
    limits: Optional[CGM] = CGM(cpu=1, mem="10M")
    # Don't specify this in the constructor, it's a derived field
    resources: Optional[dict[str, CGM]] = None

    def model_post_init(self, *args):
        # Move requests and limits to resources so they match the Pod layout.
        self.resources = dict(requests=self.requests, limits=self.limits)
        self.requests = self.limits = None


def make_node(name: str, taints: Optional[List[Taint]] = None, labels: Optional[dict] = None):
    """
    Construct a Node dict from a generic chunk of node YAML that we can alter to simulate different
    responses from the K8S API.
    :param name: Node name
    """
    node = yaml.safe_load(_resource("sample_node.yaml"))
    node["metadata"]["name"] = name
    node["metadata"]["uid"] = "uid-" + name
    if taints:
        node["spec"]["taints"] = [taint.model_dump(exclude_none=True) for taint in taints]
    if labels is not None:
        node["metadata"]["labels"] = labels
    return node


def make_pod(name: str,
             no_metadata: bool = False,
             name_at_root: bool = False,
             no_name: bool = False,
             is_daemon: bool = False,
             creation_ts: int = UNIT_TEST_TIMEBASE,
             namespace: Optional[str] = None,
             node_name: Optional[str] = None,
             containers: List[Container] = [Container()],
             labels: Optional[dict] = None,
             phase: Optional[str] = "Running",
             ):
    """
    Construct a Pod dict from a generic chunk of pod YAML that we can alter to simulate different
    responses from the K8S API.

    :param no_metadata: Pretend there is no metadata
    :param name_at_root: Put the object name at top level, not in the metadata
    :param no_name: Pretend there is no object name
    """
    obj = yaml.safe_load(_resource("sample_pod.yaml"))
    if name_at_root:
        obj["name"] = name
    elif not no_name:
        obj["metadata"]["name"] = name
    obj["metadata"]["uid"] = "uid-" + name
    if no_metadata:
        del obj["metadata"]
    if is_daemon:
        obj["metadata"]["ownerReferences"] = [{"kind": "DaemonSet"}]
    if namespace:
        obj["metadata"]["namespace"] = namespace
    if node_name:
        obj["spec"]["nodeName"] = node_name
    if labels is not None:
        obj["metadata"]["labels"] = labels
    if creation_ts and not no_metadata:
        obj["metadata"]["creationTimestamp"] = to_utc(creation_ts)
    obj["spec"]["containers"] = [c.model_dump(by_alias=True, exclude_none=True) for c in containers]
    obj["status"]["phase"] = phase
    return obj


def make_job(name: str,
             namespace: str = None,
             active_count: Optional[int] = None,
             condition: Optional[Tuple[str, str, Optional[str]]] = None,
             pod: Optional[dict] = None,
            labels: Optional[dict] = None,
             ):
    """
    Construct a Job dict from a generic chunk of pod YAML that we can alter to simulate different
    responses from the K8S API.

    :param name: Job name
    :param active_count: If present, the number of active pods
    :param condition: If present, a condition tuple (type, status, reason)
    :param: pod: If present, a pod dict to be used as the template, returned from make_pod
    """
    obj = yaml.safe_load(_resource("sample_job.yaml"))
    obj["metadata"]["name"] = name
    obj["metadata"]["uid"] = "uid-" + name
    obj["metadata"]["labels"]["job-name"] = name
    if namespace is not None:
        obj["metadata"]["namespace"] = namespace
    if active_count is not None:
        obj["status"]["active"] = active_count
    if condition is not None:
        obj["status"]["conditions"] = [{"type": condition[0], "status": condition[1], "reason": condition[2]}]
    if labels is not None:
        obj["metadata"]["labels"] = labels
    if pod is not None:
        obj["spec"]["template"]["spec"] = pod["spec"]
    return obj


def assert_query(sql: str, expected: Union[str, list], all_ns: bool = False):
    """
    Run a query in the "nocontext" namespace and compare the result with expected output.
    :param sql: SQL query
    :param expected: Output as it would be shown at the CLI.  This will be dedented so the
        caller can indent for neatness.  Or, if a list, each item will be checked in order.
    :param all_ns: FIXME temporary hack until we get namespaces out of engine.py
    """
    schema = Registry.get().get_schema("kubernetes")
    schema.impl.set_namespace(all_ns, "__all" if all_ns else "default")
    engine = Engine(schema, Settings(), "nocontext")
    if isinstance(expected, str):
        actual = engine.query_and_format(Query(sql=sql))
        assert actual.strip() == textwrap.dedent(expected).strip()
    else:
        actual, _ = engine.query(Query(sql=sql))
        assert actual == expected


def _resource(filename: str):
    # TODO: rename me
    return Path(__file__).parent.joinpath("resources", filename).read_text()



from .age import Age, parse_age, to_age
from .clock import UNIT_TEST_TIMEBASE
from .misc import (
    ConfigPath,
    debug,
    debugging,
    dprint,
    fail,
    KPath,
    kube_home,
    kugl_home,
    KuglError,
    parse_utc,
    run,
    to_utc,
    warn,
    WHITESPACE,
)
from .size import parse_size, to_size, parse_cpu
from .sqlite import SqliteDb

import kugl.util.clock as clock



import datetime as dt
import re
from typing import Dict

import funcy as fn


class Age(dt.timedelta):
    """
    A specialization of timedelta that handles age strings like "10s", "5m30s", "1h", "2d12h".
    Also, an Age is always non-negative.
    FIXME: Don't subclass timedelta, it's breaking the constructor contract.
    """

    AGE_RE = re.compile(r"(\d+[a-z])+")
    AGE_PART = re.compile(r"\d+[a-z]")

    def __new__(cls, *args, **kwargs):
        """
        Create a new Age object.  Parameters may be one of
        - a string like "10s", "5m30s", "1h", "2d12h"
        - an integer number of seconds
        - kwargs to pass to timedelta
        """
        if args:
            if kwargs:
                raise ValueError("Cannot specify both positional and keyword arguments")
            if len(args) == 3:
                # This is a call from deepcopy, bump it upstairs
                return super().__new__(cls, *args)
            if len(args) > 1:
                raise ValueError(f"Too many positional arguments: {args}")
            arg = args[0]
            if isinstance(arg, str):
                return super().__new__(cls, **Age.parse(arg))
            elif isinstance(arg, int) or isinstance(arg, float):
                return super().__new__(cls, seconds=abs(arg))
            else:
                raise ValueError(f"Invalid argument type: {arg}, {type(arg)}")
        elif not kwargs:
            raise ValueError("Must specify positional or keyword arguments")
        else:
            return super().__new__(cls, **kwargs)

    @classmethod
    def parse(cls, x: str) -> Dict[str, int]:
        """
        Convert a string like "10s", "5m30s", "1h", "2d12h" to Python timedelta dict, using suffixes to
        mean s = seconds, m = minutes, h = hours, d = days.
        """
        x = x.strip()
        if not x:
            raise ValueError("Empty argument")
        if fn.silent(int)(x) is not None:
            return {"seconds": int(x)}
        if not cls.AGE_RE.match(x):
            raise ValueError(f"Invalid age syntax: {x}")
        suffixes = {"s": "seconds", "m": "minutes", "h": "hours", "d": "days"}

        def _parse(part):
            amount, unit = int(part[:-1]), part[-1]
            if unit not in suffixes:
                raise ValueError(f"Invalid suffix {unit}, must be one of [dhms]")
            return (suffixes[unit], amount)

        return dict(_parse(part) for part in cls.AGE_PART.findall(x))

    def render(self):
        """
        Render the age as a string like "10s", "5m30s", "1h", "2d12h".
        """
        days = self.days
        if days > 9:
            return f"{days}d"
        hours = self.seconds // 3600
        if days > 1:  # kubectl prints hours up to 47
            return f"{days}d{hours}h" if hours else f"{days}d"
        if days > 0 or hours > 9:
            return f"{days * 24 + hours}h"
        minutes = (self.seconds % 3600) // 60
        if hours > 2:  # kubectl prints minutes up to 179
            return f"{hours}h{minutes}m" if minutes else f"{hours}h"
        if hours > 0 or minutes > 9:
            return f"{hours * 60 + minutes}m"
        seconds = int(self.seconds % 60)
        if minutes > 0:
            return f"{minutes}m{seconds}s" if seconds else f"{minutes}m"
        return f"{seconds}s"

    @property
    def value(self) -> int:
        return int(self.total_seconds())


def parse_age(age: str) -> int:
    return Age(age).value


def to_age(seconds: int) -> str:
    return Age(seconds).render()




"""
This is separate from engine.py for maintainability.
SQLite tables are defined and populated here.
"""

from typing import Optional, Type

import jmespath
from pydantic import Field, BaseModel

from .config import ColumnDef, ExtendTable, CreateTable

# TODO: make abstract
# TODO: completely sever from user configs
from ..util import fail, dprint, debugging


class TableDef(BaseModel):
    """
    Capture a table definition from the @table decorator, example:
        @table(schema="kubernetes", name="pods", resource="pods")
    """
    cls: Type
    name: str
    schema_name: str = Field(..., alias="schema")
    resource: str


class Table:
    """The engine-level representation of a table, independent of the config file format"""

    def __init__(self, name: str, resource: str, schema: str, extras: list[ColumnDef]):
        """
        :param name: the table name, e.g. "pods"
        :param resource: the Kubernetes resource type, e.g. "pods"
        :param schema: the SQL schema, e.g. "name TEXT, age INTEGER"
        :param extras: extra column definitions from user configs (not from Python-defined tables)
        """
        self.name = name
        self.resource = resource
        self.schema = schema
        self.extras = extras

    def build(self, db, kube_data: dict):
        """Create the table in SQLite and insert the data.

        :param db: the SqliteDb instance
        :param kube_data: the JSON data from 'kubectl get'
        """
        context = RowContext(kube_data)
        db.execute(f"CREATE TABLE {self.name} ({self.schema})")
        item_rows = list(self.make_rows(context))
        if item_rows:
            if self.extras:
                extend_row = lambda item, row: row + tuple(column.extract(item, context) for column in self.extras)
            else:
                extend_row = lambda item, row: row
            rows = [extend_row(item, row) for item, row in item_rows]
            placeholders = ", ".join("?" * len(rows[0]))
            db.execute(f"INSERT INTO {self.name} VALUES({placeholders})", rows)

    @staticmethod
    def column_schema(columns: list[ColumnDef]) -> str:
        return ", ".join(f"{c.name} {c._sqltype}" for c in columns)


class TableFromCode(Table):
    """A table created from Python code, not from a user config file."""

    def __init__(self, table_def: TableDef, extender: Optional[ExtendTable]):
        """
        :param table_def: a TableDef from the @table decorator
        :param extender: an ExtendTable object from the extend: section of a user config file
        """
        impl = table_def.cls()
        schema = impl.schema
        if extender:
            schema += ", " + Table.column_schema(extender.columns)
            extras = extender.columns
        else:
            extras = []
        dprint("schema", f"Table {table_def.name} schema: {schema}")
        super().__init__(table_def.name, table_def.resource, schema, extras)
        self.impl = impl

    def make_rows(self, context: "RowContext") -> list[tuple[dict, tuple]]:
        """Delegate to the user-defined table implementation."""
        return self.impl.make_rows(context)


class TableFromConfig(Table):
    """A table created from a create: section in a user config file, rather than in Python"""

    def __init__(self, name: str, creator: CreateTable, extender: Optional[ExtendTable]):
        """
        :param name: the table name, e.g. "pods"
        :param creator: a CreateTable object from the create: section of a user config file
        :param extender: an ExtendTable object from the extend: section of a user config file
        """
        if creator.row_source is None:
            # FIXME: kubernetes-specific
            self.itemizer = lambda data: data["items"]
        else:
            self.itemizer = lambda data: self._itemize(creator.row_source, data)
        schema = Table.column_schema(creator.columns)
        extras = creator.columns
        if extender:
            schema += ", " + Table.column_schema(extender.columns)
            extras += extender.columns
        super().__init__(name, creator.resource, schema, extras)
        self.row_source = creator.row_source

    def make_rows(self, context: "RowContext") -> list[tuple[dict, tuple]]:
        """
        Itemize the data according to the configuration, but return empty rows; all the
        columns will be added by Table.build.
        """
        if self.row_source is not None:
            items = self._itemize(self.row_source, context)
        else:
            # FIXME: this default only applies to Kubernetes
            items = context.data["items"]
        return [(item, tuple()) for item in items]

    def _itemize(self, row_source: list[str], context: "RowContext") -> list[dict]:
        """
        Given a row_source like
          row_source:
            - items
            - spec.taints
        Iterate through each level of the source spec, marking object parents, and generating
        successive row values
        """
        items = [context.data]
        debug = debugging("itemize")
        for source in row_source:
            try:
                finder = jmespath.compile(source)
            except jmespath.exceptions.ParseError as e:
                fail(f"invalid row_source {source} for {self.name} table", e)
            new_items = []
            if debug:
                print(f"Itemizing {self.name} at {source} got {len(items)} hits")
            for item in items:
                found = finder.search(item)
                if isinstance(found, list):
                    for child in found:
                        context.set_parent(child, item)
                        new_items.append(child)
                elif found is not None:
                    context.set_parent(found, item)
                    new_items.append(found)
            items = new_items
        return items


class RowContext:
    """Provide helpers to row-generating functions.

    Primarily, the `.data` attribute holds the JSON data from 'kubectl get' or similar.
    The `.set_parent` and `.get_parent` methods allow row-generating functions to track
    parent objects as they iterate through nested data structures."""

    def __init__(self, data):
        self.data = data
        self.debug = debugging("extract")
        self._parents = {}

    def set_parent(self, child, parent):
        self._parents[id(child)] = parent

    def get_parent(self, child, depth: int = 1):
        while depth > 0 and child is not None:
            child = self._parents.get(id(child))
            depth -= 1
        return child

    def get_root(self, child):
        while (parent := self._parents.get(id(child))) is not None:
            child = parent
        return child




"""
Assorted query tests not covered elsewhere.
"""
import io
import json
import sys
import textwrap

import pytest

from kugl.main import main1
from kugl.util import KuglError
from .testing import kubectl_response, assert_query


def test_non_sql_types(test_home):
    (test_home / "kubernetes.yaml").write_text("""
      resources:
        - name: things
          namespaced: false
      create:
        - table: things
          resource: things
          columns:
            - name: size
              type: size
              path: size
            - name: cpu
              type: cpu
              path: cpu
            - name: age
              type: age
              path: age
            - name: date
              type: date
              path: date
    """)
    kubectl_response("things", {
        "items": [
            {"size": "10Ki", "cpu": "2.5", "age": "2d", "date": "2021-01-01"},
            {"size": "2Gi", "cpu": "300m", "age": "4h", "date": "2021-12-31T23:59:59Z"},
        ]
    })
    assert_query("SELECT to_size(size) AS s, cpu, to_age(age) AS a, to_utc(date) AS d FROM things ORDER BY 1", """
        s        cpu  a    d
        10Ki     2.5  2d   2021-01-01T00:00:00Z
        2.0Gi    0.3  4h   2021-12-31T23:59:59Z
    """)


def test_too_many_parents(test_home):
    (test_home / "kubernetes.yaml").write_text("""
      resources:
        - name: things
      create:
        - table: things
          resource: things
          columns:
            - name: something
              path: ^^^invalid
    """)
    kubectl_response("things", {
        "items": [
            {"something": "foo"},
            {"something": "foo"},
        ]
    })
    with pytest.raises(KuglError, match="Missing parent or too many . while evaluating ...invalid"):
        assert_query("SELECT * FROM things", "")


def test_config_with_missing_resource(test_home):
    (test_home / "kubernetes.yaml").write_text("""
        create:
          - table: stuff
            resource: stuff
            columns: []
    """)
    with pytest.raises(KuglError, match="Table 'stuff' needs unknown resource 'stuff'"):
        assert_query("SELECT * FROM stuff", "")


def test_select_from_stdin(test_home, monkeypatch, capsys):
    (test_home / "hr.yaml").write_text("""
        resources:
          - name: stdin
            file: stdin
        create:
          - table: people
            resource: stdin
            row_source:
              - people
            columns:
              - name: name
                path: name
              - name: age
                path: age
                type: integer
    """)
    monkeypatch.setattr(sys, "stdin", io.StringIO(json.dumps({"people": [
        {"name": "Jim", "age": 42},
        {"name": "Jill", "age": 43},
    ]})))
    main1(["SELECT name, age FROM hr.people"])
    out, _ = capsys.readouterr()
    assert out.strip() == textwrap.dedent("""
        name      age
        Jim        42
        Jill       43
    """).strip()




"""
Tests for the nodes and taints tables.
"""

import yaml

from kugl.impl.config import parse_model, UserConfig
from kugl.util import fail
from .testing import make_node, kubectl_response, assert_query, Taint


def test_node_query(test_home):
    (test_home / "kubernetes.yaml").write_text("""
        extend:
          - table: nodes
            columns:
              - name: instance_type
                label:
                  - node.kubernetes.io/instance-type
                  - beta.kubernetes.io/instance-type
            
    """)
    kubectl_response("nodes", {
        "items": [
            make_node("node-1", labels={"node.kubernetes.io/instance-type": "a40"}),
            make_node("node-2", labels={"beta.kubernetes.io/instance-type": "a40"}),
        ]
    })
    assert_query("SELECT * FROM nodes ORDER BY name", """
        name    uid           cpu_alloc    gpu_alloc     mem_alloc    cpu_cap    gpu_cap       mem_cap  instance_type
        node-1  uid-node-1           93            4  807771639808         96          4  810023981056  a40
        node-2  uid-node-2           93            4  807771639808         96          4  810023981056  a40
    """)


def test_taint_query(test_home):
    kubectl_response("nodes", {
        "items": [
            make_node("node-1"),
            make_node("node-2", taints=[Taint(key="node.kubernetes.io/unschedulable", effect="NoSchedule"),
                                        Taint(key="node.kubernetes.io/unreachable", effect="NoExecute")
                                        ]),
            make_node("node-3", taints=[Taint(key="mycompany.com/priority", effect="NoSchedule", value="true")
                                        ]),
        ]
    })
    assert_query("""
        SELECT n.name, nt.key, nt.effect
        FROM nodes n join node_taints nt on nt.node_uid = n.uid
        ORDER BY 1, 2
    """, """
        name    key                               effect
        node-2  node.kubernetes.io/unreachable    NoExecute
        node-2  node.kubernetes.io/unschedulable  NoSchedule
        node-3  mycompany.com/priority            NoSchedule
    """)


def test_node_labels(test_home):
    kubectl_response("nodes", {
        "items": [
            make_node("node-1", labels=dict(foo="bar")),
            make_node("node-2", labels=dict(a="b", c="d", e="f")),
            make_node("node-3", labels=dict()),
            make_node("node-4", labels=dict(one="two", three="four")),
        ]
    })
    assert_query("SELECT node_uid, key, value FROM node_labels ORDER BY 2, 1", """
        node_uid    key    value
        uid-node-2  a      b
        uid-node-2  c      d
        uid-node-2  e      f
        uid-node-1  foo    bar
        uid-node-4  one    two
        uid-node-4  three  four
    """)



"""
A wrapper on the system clock that can be replaced for unit testing.
"""

import time
from abc import abstractmethod
from typing import Optional

# For use with simulated clock
UNIT_TEST_TIMEBASE = 1733798942


class Clock:

    @abstractmethod
    def set(self, epoch: int):
        ...

    @abstractmethod
    def now(self) -> int:
        ...

    @abstractmethod
    def sleep(self, seconds: int):
        ...

    @property
    @abstractmethod
    def is_simulated(self) -> bool:
        ...


class RealClock(Clock):

    def set(self, epoch: int):
        pass

    def now(self) -> int:
        return int(time.time())

    def sleep(self, seconds: int):
        time.sleep(seconds)

    @property
    def is_simulated(self) -> bool:
        return False


CLOCK = RealClock()


def simulate_time():
    global CLOCK
    CLOCK = FakeClock()


class FakeClock(Clock):

    def __init__(self, epoch: Optional[int] = None):
        self._epoch = epoch or int(time.time())

    def set(self, epoch: int):
        self._epoch = epoch

    def now(self) -> int:
        return self._epoch

    def sleep(self, seconds: int):
        self._epoch += seconds

    @property
    def is_simulated(self) -> bool:
        return True



import os
from pathlib import Path

import pytest

from kugl.util import UNIT_TEST_TIMEBASE, kube_home, clock, KPath

# Add tests/ folder to $PATH so running 'kubectl ...' invokes our mock, not the real kubectl.
os.environ["PATH"] = f"{Path(__file__).parent}:{os.environ['PATH']}"

# Some behaviors have to change in tests, sorry
os.environ["KUGL_UNIT_TESTING"] = "true"


def pytest_sessionstart(session):
    # Tell Pytest where there are assertions in files that aren't named "test_*"
    pytest.register_assert_rewrite("tests.testing")
    # Use a clock we can control, in place of system time.
    clock.simulate_time()
    clock.CLOCK.set(UNIT_TEST_TIMEBASE)


@pytest.fixture(scope="function")
def test_home(tmp_path, monkeypatch):
    monkeypatch.setenv("KUGL_HOME", str(tmp_path))
    monkeypatch.setenv("KUGL_MOCKDIR", str(tmp_path / "cache"))
    kube_home().mkdir()
    kube_home().joinpath("config").write_text("current-context: nocontext")
    yield KPath(tmp_path)


"""
Tests for the jobs table.
"""

from .testing import make_job, kubectl_response, assert_query


def test_job_status(test_home):
    kubectl_response("jobs", {
        "items": [
            make_job("job-1"),
            make_job("job-2", active_count=1),
            make_job("job-3", namespace="xyz", condition=("Failed", "False", None)),
            make_job("job-4", namespace="xyz", condition=("Failed", "True", None)),
            make_job("job-5", condition=("Failed", "True", "DeadlineExceeded")),
            make_job("job-6", condition=("Suspended", "True", None)),
            make_job("job-7", condition=("Complete", "True", None)),
            make_job("job-8", condition=("FailureTarget", "False", None)),
            make_job("job-9", condition=("SuccessCriteriaMet", "False", None)),
        ]
    })
    assert_query("SELECT name, uid, namespace, status FROM jobs ORDER BY 1", """
        name    uid        namespace    status
        job-1   uid-job-1  example      Unknown
        job-2   uid-job-2  example      Running
        job-3   uid-job-3  xyz          Unknown
        job-4   uid-job-4  xyz          Failed
        job-5   uid-job-5  example      DeadlineExceeded
        job-6   uid-job-6  example      Suspended
        job-7   uid-job-7  example      Complete
        job-8   uid-job-8  example      Failed
        job-9   uid-job-9  example      Complete
    """)


def test_job_labels(test_home):
    kubectl_response("jobs", {
        "items": [
            make_job("job-1", labels=dict(foo="bar")),
            make_job("job-2", labels=dict(a="b", c="d", e="f")),
            make_job("job-3", labels=dict()),
            make_job("job-4", labels=dict(one="two", three="four")),
        ]
    })
    assert_query("SELECT job_uid, key, value FROM job_labels ORDER BY 2, 1", """
        job_uid    key    value
        uid-job-2  a      b
        uid-job-2  c      d
        uid-job-2  e      f
        uid-job-1  foo    bar
        uid-job-4  one    two
        uid-job-4  three  four
    """)



"""
Tests for user configuration file content.
"""
import pytest

from kugl.impl.config import Settings, UserConfig, parse_model, ExtendTable, CreateTable, UserInit

import yaml

from kugl.util import Age, KuglError


def test_settings_defaults():
    s = Settings()
    assert s.cache_timeout == Age(120)
    assert s.reckless == False


def test_settings_custom():
    s = Settings(cache_timeout=Age(5), reckless=True)
    assert s.cache_timeout == Age(5)
    assert s.reckless == True


def test_empty_config():
    c = UserConfig()
    assert c.extend == []
    assert c.create == []


def test_empty_init():
    c = UserInit()
    assert c.settings.cache_timeout == Age(120)
    assert c.settings.reckless == False
    assert c.shortcuts == {}


def test_config_with_table_extension():
    c, e = parse_model(UserConfig, yaml.safe_load("""
        extend:
          - table: pods
            columns:
              - name: foo
                path: metadata.name
              - name: bar
                type: integer
                path: metadata.creationTimestamp
    """))
    assert e is None
    columns = c.extend[0].columns
    assert columns[0].name == "foo"
    assert columns[0].type == "text"
    assert columns[0].path == "metadata.name"
    assert columns[1].name == "bar"
    assert columns[1].type == "integer"
    assert columns[1].path == "metadata.creationTimestamp"


def test_config_with_table_creation():
    c, e = parse_model(UserConfig, yaml.safe_load("""
        resources:
          - name: pods
        create:
          - table: pods
            resource: pods
            columns:
              - name: foo
                path: metadata.name
              - name: bar
                type: integer
                path: metadata.creationTimestamp
    """))
    assert e is None
    pods = c.create[0]
    assert pods.resource == "pods"
    columns = pods.columns
    assert columns[0].name == "foo"
    assert columns[0].type == "text"
    assert columns[0].path == "metadata.name"
    assert columns[1].name == "bar"
    assert columns[1].type == "integer"
    assert columns[1].path == "metadata.creationTimestamp"


def test_unknown_type():
    _, errors = parse_model(ExtendTable, yaml.safe_load("""
        table: xyz
        columns:
          - name: foo
            type: unknown_type
            path: metadata.name
    """))
    assert len(errors) == 1
    assert "columns.0.type: Input should be" in errors[0]


def test_missing_fields_for_create():
    _, errors = parse_model(CreateTable, yaml.safe_load("""
        table: xyz
        columns:
          - name: foo
            path: metadata.name
    """))
    assert set(errors) == set([
        "resource: Field required",
    ])


def test_unexpected_keys():
    _, errors = parse_model(ExtendTable, yaml.safe_load("""
        table: xyz
        columns:
          - name: foo
            path: metadata.name
            unexpected: 42
    """))
    assert errors == ["columns.0.unexpected: Extra inputs are not permitted"]


def test_invalid_jmespath():
    _, errors = parse_model(ExtendTable, yaml.safe_load("""
        table: xyz
        columns:
          - name: foo
            path: ...name
    """))
    assert errors == ["columns.0: Value error, invalid JMESPath expression ...name in column foo"]


def test_cannot_have_both_path_and_label():
    _, errors = parse_model(ExtendTable, yaml.safe_load("""
        table: xyz
        columns:
          - name: foo
            type: text
            path: xyz
            label: xyz
    """))
    assert errors == ["columns.0: Value error, cannot specify both path and label"]


def test_must_specify_path_or_label():
    _, errors = parse_model(ExtendTable, yaml.safe_load("""
        table: xyz
        columns:
          - name: foo
    """))
    assert errors == ["columns.0: Value error, must specify either path or label"]


"""
Assorted utility functions / classes with no obvious home.
"""
import json
import os
import re
import subprocess as sp
import sys
from pathlib import Path
from typing import Optional, Union

import arrow
import yaml

from kugl.util import Age, clock

WHITESPACE = re.compile(r"\s+")
DEBUG_FLAGS = {}


def run(args: Union[str, list[str]], error_ok=False):
    """
    Invoke an external command, which may be a list or a string; in the latter case it will be
    interpreted using bash -c.  Returns exit status, stdout and stderr.
    """
    if isinstance(args, str):
        args = ["bash", "-c", args]
    dprint("fetch", f"Running {' '.join(args)}")
    p = sp.run(args, stdout=sp.PIPE, stderr=sp.PIPE, encoding="utf-8")
    if p.returncode != 0 and not error_ok:
        print(f"Failed to run [{' '.join(args)}]:", file=sys.stderr)
        print(p.stderr, file=sys.stderr, end="")
        sys.exit(p.returncode)
    return p.returncode, p.stdout, p.stderr


def parse_utc(utc_str: str) -> int:
    return arrow.get(utc_str).int_timestamp


def to_utc(epoch: int) -> str:
    return arrow.get(epoch).to('utc').format('YYYY-MM-DDTHH:mm:ss') + 'Z'


def warn(message: str):
    print(message, file=sys.stderr)


def fail(message: str, e: Optional[Exception] = None):
    if e is not None:
        raise KuglError(message) from e
    raise KuglError(message)


class KuglError(Exception):
    pass


def debug(features: list[str], on: bool = True):
    """Turn debugging on or off for a set of features.

    :param features: list of feature names, parsed from the --debug command line option;
        "all" means everything.
    """
    for feature in features:
        if feature == "all" and not on:
            DEBUG_FLAGS.clear()
        else:
            DEBUG_FLAGS[feature] = on


def debugging(feature: str = None) -> bool:
    """Check if a feature is being debugged."""
    if feature is None:
        return len(DEBUG_FLAGS) > 0
    return DEBUG_FLAGS.get(feature) or DEBUG_FLAGS.get("all")


def dprint(feature, *args, **kwargs):
    """Print a debug message if the given feature is being debugged."""
    if debugging(feature):
        print(*args, **kwargs)


class KPath(type(Path())):
    """It would be nice if Path were smarter, so do that."""

    def is_world_writeable(self) -> bool:
        return self.stat().st_mode & 0o2 == 0o2

    def parse_json(self):
        return json.loads(self.read_text())

    def parse_yaml(self):
        return yaml.safe_load(self.read_text())

    def set_age(self, age: Age):
        time = clock.CLOCK.now() - age.value
        os.utime(str(self), times=(time, time))


class ConfigPath(KPath):
    """Same as a KPath but adds debug statements"""

    def parse_json(self):
        dprint("config", f"Loading {self}")
        return super().parse_json()

    def parse_yaml(self):
        dprint("config", f"Loading {self}")
        return super().parse_yaml()


def kugl_home() -> KPath:
    # KUGL_HOME override is for unit tests, not for users (as least for now)
    if "KUGL_HOME" in os.environ:
        return KPath(os.environ["KUGL_HOME"])
    return KPath.home() / ".kugl"


def kube_home() -> KPath:
    # KUGL_HOME override is for unit tests, not for users (as least for now)
    if "KUGL_HOME" in os.environ:
        return KPath(os.environ["KUGL_HOME"]) / ".kube"
    return KPath.home() / ".kube"


"""
More assorted tests, should these be combined with test_misc.py?
"""

import jmespath
import pytest

from kugl.util import Age, parse_size, to_size, dprint, debug, parse_cpu


@pytest.mark.parametrize("input_args,input_kwargs,expected", [
    ([], {}, ValueError("Must specify positional or keyword arguments")),
    ([1], {"x":1}, ValueError("Cannot specify both positional and keyword arguments")),
    ([1, 1], {}, ValueError("Too many positional arguments")),
    ([""], {}, ValueError("Empty argument")),
    (["xxx"], {}, ValueError("Invalid age syntax: xxx")),
    (["1x"], {}, ValueError("Invalid suffix x, must be one of")),
    ([], {"seconds": 10}, "10s"),
    ([], {"minutes": 5}, "5m"),
    ([], {"hours": 3}, "3h"),
    ([], {"days": 2}, "2d"),
    ([2], {}, "2s"),
    ([2.5], {}, "2s"),
    ([-2], {}, "2s"),
    (["10d9h"], {}, "10d"),
    (["9d9h"], {}, "9d9h"),
    (["9d"], {}, "9d"),
    (["50h"], {}, "2d2h"),
    (["1d9h"], {}, "33h"),
    (["10h40m"], {}, "10h"),
    (["9h40m"], {}, "9h40m"),
    (["9h40m20s"], {}, "9h40m"),
    (["9h20s"], {}, "9h"),
    (["3h"], {}, "3h"),
    (["2h50m"], {}, "170m"),
    (["1h"], {}, "60m"),
    (["1h20s"], {}, "60m"),
    (["80s"], {}, "1m20s"),
    (["10m20s"], {}, "10m"),
    (["9m20s"], {}, "9m20s"),
    (["8m80s"], {}, "9m20s"),
    (["0m40s"], {}, "40s"),
    (["2m"], {}, "2m"),
    (["30s"], {}, "30s"),
    (["0m"], {}, "0s"),
    (["0s"], {}, "0s"),
])
def test_age(input_args, input_kwargs, expected):
    if isinstance(expected, Exception):
        with pytest.raises(expected.__class__, match=str(expected)):
            Age(*input_args, **input_kwargs)
    else:
        assert Age(*input_args, **input_kwargs).render() == expected


@pytest.mark.parametrize("size_str, expected_result", [
    ("", "Can't translate '' to bytes"),
    ("1n", "Can't translate '1n' to bytes"),
    ("15", 15.0),
    ("15K", 15.0 * 10 ** 3),
    ("15Ki", 15.0 * 2 ** 10),
    ("15M", 15.0 * 10 ** 6),
    ("15Mi", 15.0 * 2 ** 20),
    ("15G", 15.0 * 10 ** 9),
    ("15Gi", 15.0 * 2 ** 30),
])
def test_parse_size(size_str, expected_result):
    if isinstance(expected_result, str):
        with pytest.raises(ValueError, match=expected_result):
            parse_size(size_str)
    else:
        assert parse_size(size_str) == expected_result


@pytest.mark.parametrize("cpu_str, expected_result", [
    (None, None),
    (2, 2),
    (2.0, 2.0),
    ("1.5", 1.5),
    ("300m", 0.3),
    ("50x", "could not convert string to float"),
])
def test_parse_cpu(cpu_str, expected_result):
    if isinstance(expected_result, str):
        with pytest.raises(ValueError, match=expected_result):
            parse_cpu(cpu_str)
    else:
        assert parse_cpu(cpu_str) == expected_result


@pytest.mark.parametrize("args,result", [
    ((0,), "0"),
    ((1,), "1"),
    ((999,), "999"),
    ((1023, True), "1023"),
    ((1000,), "1.0K"),
    ((1024, True), "1.0Ki"),
    ((1000 ** 2 - 1000,), "999K"),
    ((1024 ** 2 - 1024, True), "1023Ki"),
    ((1000 ** 2,), "1.0M"),
    ((1024 ** 2, True), "1.0Mi"),
    ((1000 ** 3 - 1000 ** 2,), "999M"),
    ((1024 ** 3 - 1024 ** 2, True), "1023Mi"),
    ((1000 ** 3,), "1.0G"),
    ((1024 ** 3, True), "1.0Gi"),
    ((1000 ** 4 - 1000 ** 3,), "999G"),
    ((1024 ** 4 - 1024 ** 3, True), "1023Gi"),
    ((1000 ** 4,), "1.0T"),
    ((1024 ** 4, True), "1.0Ti"),
])
def test_to_size(args: tuple, result: str):
    assert to_size(*args) == result


def test_jmespath_performance():
    """
    JMESPath performance regression test.  We use JMESPath to filter and transform
    the data returned by the Kubernetes API.
    """
    path = jmespath.compile("pods[?status.phase == 'Running'].metadata.name")
    data = {
        "pods": [
            {"status": {"phase": "Running"}, "metadata": {"name": "pod-1"}},
            {"status": {"phase": "Running"}, "metadata": {"name": "pod-2"}},
            {"status": {"phase": "Pending"}, "metadata": {"name": "pod-3"}},
        ]
    }
    # Use true system time
    import time
    start = time.time()
    for _ in range(10000):
        result = path.search({**data})
    end = time.time()
    assert end - start < 1.0
    assert result == ["pod-1", "pod-2"]


def test_dprint(capsys):
    FEATURE = "afeature"
    dprint(FEATURE, "hello")
    assert capsys.readouterr().out == ""
    debug([FEATURE])
    dprint(FEATURE, "hello")
    assert capsys.readouterr().out == "hello\n"
    debug([FEATURE], False)
    dprint(FEATURE, "hello")
    assert capsys.readouterr().out == ""



import re
from typing import Union

SIZE_RE = re.compile(r"([0-9\.]+)(([KMGTP]i?)?)")
SIZE_MULTIPLIERS = dict(K=10**3, M=10**6, G=10**9, T=10**12,
                        Ki=2**10, Mi=2**20, Gi=2**30, Ti=2**40)


def parse_size(x: Union[str, int, None]):
    """
    Translate a string a la 10K, 5Mb, 3Gi to # of bytes.  Returns an int if the result
    can be represented as an int, else a float.
    """
    if x is None:
        return None
    if isinstance(x, int):
        return x
    m = SIZE_RE.fullmatch(x)
    if m is None:
        raise ValueError(f"Can't translate '{x}' to bytes")
    amount, suffix = m.group(1), m.group(2)
    amount = float(amount) if "." in amount else int(amount)
    if suffix == "":
        return amount
    multiplier = SIZE_MULTIPLIERS.get(suffix)
    if multiplier is None:
        raise ValueError(f"Unknown size suffix in '{x}'")
    return int(amount * multiplier)


def to_size(nbytes: int, iec=False):
    """
    Given a byte count, render it as a string in the most appropriate units, suffixed by KB, MB, GB, etc.
    Larger sizes will use the appropriate unit.  The result may have a maximum of one digit after the
    decimal point.  If iec is True, use IEC binary units (KiB, MiB, etc).
    """
    one_k = 1024 if iec else 1000
    bytes = "i" if iec else ""
    if nbytes < one_k:
        return f"{nbytes}"
    elif nbytes < one_k ** 2:
        size, suffix = nbytes / one_k, "K" + bytes
    elif nbytes < one_k ** 3:
        size, suffix = nbytes / one_k ** 2, "M" + bytes
    elif nbytes < one_k ** 4:
        size, suffix = nbytes / one_k ** 3, "G" + bytes
    else:
        size, suffix = nbytes / one_k ** 4, "T" + bytes
    if size < 10:
        return f"{size:.1f}{suffix}"
    else:
        return f"{round(size)}{suffix}"


def parse_cpu(x: Union[str, float, int, None]) -> float:
    """
    Translate a "number of CPUs" field like "2", "1.5", or "300m" to float.
    """
    if x is None or isinstance(x, float) or isinstance(x, int):
        return x
    if x.endswith("m"):
        return float(x[:-1]) / 1000
    return float(x)



"""
Command-line entry point.
"""

import os
from argparse import ArgumentParser
import sys
from sqlite3 import DatabaseError
from typing import List, Optional, Union

import yaml

from kugl.impl.registry import Registry
from kugl.impl.engine import Engine, Query, CHECK, NEVER_UPDATE, ALWAYS_UPDATE
from kugl.impl.config import UserConfig, UserInit, parse_file
from kugl.util import Age, fail, debug, debugging, kugl_home, kube_home, ConfigPath, dprint, KuglError


def main() -> None:
    main1(sys.argv[1:])


def main1(argv: List[str]):

    if "KUGL_UNIT_TESTING" in os.environ and "KUGL_MOCKDIR" not in os.environ:
        # Never enter main in tests unless test_home fixture is in use, else we could read
        # the user's init file.
        sys.exit("Unit test state error")

    try:
        return main2(argv)
    except KuglError as e:
        # These are raised by fail(), we only want the error message.
        severe, exc = False, e
    except DatabaseError as e:
        # DB errors are common when writing queries, don't make them look like a crash.
        severe, exc = False, e
    except Exception as e:
        severe, exc = True, e
    if severe or debugging() or "KUGL_UNIT_TESTING" in os.environ:
        raise exc
    print(exc, file=sys.stderr)
    sys.exit(1)


def main2(argv: List[str]):

    kugl_home().mkdir(exist_ok=True)
    if not argv:
        fail("missing sql query")

    # Load init file.
    init_file = ConfigPath(kugl_home() / "init.yaml")
    init, errors = parse_file(UserInit, init_file)
    if errors:
        fail("\n".join(errors))

    # Check for shortcuts now, because they can include command line options.  But the
    # command line as given also applies.  So we have to treat the last arg as SQL or
    # shortcut name, even before we see the options.
    if " " not in argv[-1]:
        if not (new_argv := init.shortcuts.get(argv[-1])):
            fail(f"No shortcut named '{argv[-1]}' is defined in ~/.kugl/init.yaml")
        return main1(argv[:-1] + new_argv)

    # Need the query schema for command line parsing.
    # FIXME: Move the namespace & cache flag out of the query
    rgy = Registry.get()
    query = Query(sql=argv[-1])
    schema_refs = {ref.schema_name for ref in query.table_refs}
    if len(schema_refs) == 0:
        schema = rgy.get_schema("empty")
    elif len(schema_refs) == 1:
        schema = rgy.get_schema(next(iter(schema_refs)))
    else:
        fail("Cross-schema query not implemented yet")

    ap = ArgumentParser()
    schema.impl.add_cli_options(ap)
    ap.add_argument("-D", "--debug", type=str)
    ap.add_argument("-c", "--cache", default=False, action="store_true")
    ap.add_argument("-r", "--reckless", default=False, action="store_true")
    ap.add_argument("-t", "--timeout", type=str)
    ap.add_argument("-u", "--update", default=False, action="store_true")
    ap.add_argument("sql")
    args = ap.parse_args(argv)

    schema.impl.handle_cli_options(args)
    if args.cache and args.update:
        fail("Cannot use both -c/--cache and -u/--update")

    cache_flag = ALWAYS_UPDATE if args.update else NEVER_UPDATE if args.cache else CHECK
    if args.debug:
        debug(args.debug.split(","))
    if args.reckless:
        init.settings.reckless = True
    if args.timeout:
        init.settings.cache_timeout = Age(args.timeout)
    dprint("init", f"Settings: {init.settings}")

    kube_config = kube_home() / "config"
    if not kube_config.exists():
        fail(f"Missing {kube_config}, can't determine current context")

    current_context = (yaml.safe_load(kube_config.read_text()) or {}).get("current-context")
    if not current_context:
        fail("No current context, please run kubectl config use-context ...")

    engine = Engine(schema, init.settings, current_context)
    # FIXME bad reference to namespace
    sql = query.sql_schemaless
    print(engine.query_and_format(Query(sql=sql, namespace=schema.impl.ns, cache_flag=cache_flag)))


if __name__ == "__main__":
    main()


"""
Tests for data cache timeout behavior.
"""

from kugl.impl.config import ResourceDef
from kugl.impl.engine import DataCache, CHECK, NEVER_UPDATE, ALWAYS_UPDATE
from kugl.util import Age


def test_cache(test_home):
    NS = "default"
    cache = DataCache(test_home, Age("1m"))

    pods = ResourceDef(name="pods")
    jobs = ResourceDef(name="jobs")
    nodes = ResourceDef(name="nodes", namespaced=False)
    events = ResourceDef(name="events", cacheable=False)
    all = {pods, jobs, nodes, events}

    pods_file = cache.cache_path(NS, "pods")
    jobs_file = cache.cache_path(NS, "jobs")
    nodes_file = cache.cache_path(NS, "nodes")
    events_file = cache.cache_path(NS, "events")

    # Note: don't write jobs data
    pods_file.write_text("{}")
    nodes_file.write_text("{}")
    events_file.write_text("{}")

    pods_file.set_age(Age("50s"))  # not expired
    nodes_file.set_age(Age("70s"))  # expired
    events_file.set_age(Age("50s"))  # not expired, but not cacheable

    refresh, max_age = cache.advise_refresh(NS, all, NEVER_UPDATE)
    assert refresh == {jobs, events}
    assert max_age == 70

    refresh, max_age = cache.advise_refresh(NS, all, CHECK)
    assert refresh == {jobs, nodes, events}
    assert max_age == 50

    refresh, max_age = cache.advise_refresh(NS, all, ALWAYS_UPDATE)
    assert refresh == all
    assert max_age is None




"""
Assorted tests for various edge cases and error conditions.
Some of these are just to achieve 100% coverage.
"""

import os
from pathlib import Path

import pytest

from kugl.builtins.helpers import Limits
from kugl.main import main1
from kugl.util import Age, KuglError, kube_home, kugl_home


def test_no_resources():
    assert Limits.extract(None) == Limits(None, None, None)


def test_kube_home_missing(test_home, tmp_path):
    os.environ["KUGL_HOME"] = str(tmp_path / "doesnt_exist")
    with pytest.raises(KuglError, match="can't determine current context"):
        main1(["select 1"])


def test_no_kube_context(test_home, tmp_path):
    kube_home().joinpath("config").write_text("")
    with pytest.raises(KuglError, match="No current context"):
        main1(["select 1"])


def test_enforce_mockdir(test_home, monkeypatch):
    monkeypatch.delenv("KUGL_MOCKDIR")
    with pytest.raises(SystemExit, match="Unit test state error"):
        main1(["select 1"])


def test_kube_home_without_envar(monkeypatch):
    monkeypatch.setenv("KUGL_HOME", "xxx")  # must exist before deleting
    monkeypatch.delenv("KUGL_HOME")
    assert kube_home() == Path.home() / ".kube"


def test_kugl_home_without_envar(monkeypatch):
    monkeypatch.setenv("KUGL_HOME", "xxx")  # must exist before deleting
    monkeypatch.delenv("KUGL_HOME")
    assert kugl_home() == Path.home() / ".kugl"


def test_reject_world_writeable_config(test_home):
    init_file = kugl_home() / "init.yaml"
    init_file.write_text("foo: bar")
    init_file.chmod(0o777)
    with pytest.raises(KuglError, match="is world writeable"):
        main1(["select 1"])


@pytest.mark.skip  # FIXME re-enable without return_config hack
def test_cli_args_override_settings(test_home):
    init, _ = main1(["select 1"], return_config=True)
    assert init.settings.cache_timeout == Age(120)
    assert init.settings.reckless == False
    init, _ = main1(["-t 5", "-r", "select 1"], return_config=True)
    assert init.settings.cache_timeout == Age(5)
    assert init.settings.reckless == True


"""
Built-in table definitions for Kubernetes.

NOTE: This is not a good example of how to write user-defined tables.
FIXME: Remove references to non-API imports.
FIXME: Don't use ArgumentParser in the API.
"""
import json
from argparse import ArgumentParser
from threading import Thread

from .helpers import Limits, ItemHelper, PodHelper, JobHelper
from kugl.api import schema, table, fail
from kugl.util import parse_utc, run, WHITESPACE


@schema("kubernetes")
class KubernetesData:  # FIXME: this should be a resource type, not a schema

    def add_cli_options(self, ap: ArgumentParser):
        ap.add_argument("-a", "--all-namespaces", default=False, action="store_true")
        ap.add_argument("-n", "--namespace", type=str)

    def handle_cli_options(self, args):
        if args.all_namespaces and args.namespace:
            fail("Cannot use both -a/--all-namespaces and -n/--namespace")
        self.set_namespace(args.all_namespaces, args.namespace)

    def set_namespace(self, all_namespaces: bool, namespace: str):
        if all_namespaces:
            # FIXME: engine.py and testing.py still use this
            self.ns = "__all"
            self.all_ns = True
        else:
            self.ns = namespace or "default"
            self.all_ns = False

    def get_objects(self, kind: str, namespaced: bool)-> dict:
        """Fetch resources from Kubernetes using kubectl.

        :param kind: Kubernetes resource type e.g. "pods"
        :return: JSON as output by "kubectl get {kind} -o json"
        """
        namespace_flag = ["--all-namespaces"] if self.all_ns else ["-n", self.ns]
        if kind == "pods":
            pod_statuses = {}
            # Kick off a thread to get pod statuses
            def _fetch():
                _, output, _ = run(["kubectl", "get", "pods", *namespace_flag])
                pod_statuses.update(self._pod_status_from_pod_list(output))
            status_thread = Thread(target=_fetch, daemon=True)
            status_thread.start()
        if namespaced:
            _, output, _= run(["kubectl", "get", kind, *namespace_flag, "-o", "json"])
        else:
            _, output, _ = run(["kubectl", "get", kind, "-o", "json"])
        data = json.loads(output)
        if kind == "pods":
            # Add pod status to pods
            status_thread.join()
            def pod_with_updated_status(pod):
                metadata = pod["metadata"]
                status = pod_statuses.get(f"{metadata['namespace']}/{metadata['name']}")
                if status:
                    pod["kubectl_status"] = status
                    return pod
                return None
            data["items"] = list(filter(None, map(pod_with_updated_status, data["items"])))
        return data

    def _pod_status_from_pod_list(self, output) -> dict[str, str]:
        """
        Convert the tabular output of 'kubectl get pods' to JSON.
        :return: a dict mapping "namespace/name" to status
        """
        rows = [WHITESPACE.split(line.strip()) for line in output.strip().split("\n")]
        if len(rows) < 2:
            return {}
        header, rows = rows[0], rows[1:]
        name_index = header.index("NAME")
        status_index = header.index("STATUS")
        # It would be nice if 'kubectl get pods' printed the UID, but it doesn't, so use
        # "namespace/name" as the key.  (Can't use a tuple since this has to be JSON-dumped.)
        if self.all_ns:
            namespace_index = header.index("NAMESPACE")
            return {f"{row[namespace_index]}/{row[name_index]}": row[status_index] for row in rows}
        else:
            return {f"{self.ns}/{row[name_index]}": row[status_index] for row in rows}


@table(schema="kubernetes", name="nodes", resource="nodes")
class NodesTable:

    @property
    def schema(self):
        return """
            name TEXT,
            uid TEXT,
            cpu_alloc REAL,
            gpu_alloc REAL,
            mem_alloc INTEGER,
            cpu_cap REAL,
            gpu_cap REAL,
            mem_cap INTEGER
        """

    def make_rows(self, context) -> list[tuple[dict, tuple]]:
        for item in context.data["items"]:
            node = ItemHelper(item)
            yield item, (
                node.name,
                node.metadata.get("uid"),
                *Limits.extract(node["status"]["allocatable"], debug=context.debug).as_tuple(),
                *Limits.extract(node["status"]["capacity"], debug=context.debug).as_tuple(),
            )


@table(schema="kubernetes", name="pods", resource="pods")
class PodsTable:

    @property
    def schema(self):
        return """
            name TEXT,
            uid TEXT,
            is_daemon INTEGER,
            namespace TEXT,
            node_name TEXT,
            creation_ts INTEGER,
            command TEXT,
            phase TEXT,
            status TEXT,
            cpu_req REAL,
            gpu_req REAL,
            mem_req INTEGER,
            cpu_lim REAL,
            gpu_lim REAL,
            mem_lim INTEGER
        """

    def make_rows(self, context) -> list[tuple[dict, tuple]]:
        for item in context.data["items"]:
            pod = PodHelper(item)
            yield item, (
                pod.name,
                pod.metadata.get("uid"),
                1 if pod.is_daemon else 0,
                pod.namespace,
                pod["spec"].get("nodeName"),
                parse_utc(pod.metadata["creationTimestamp"]),
                pod.command,
                pod["status"]["phase"],
                pod["kubectl_status"],
                *pod.resources("requests", debug=context.debug).as_tuple(),
                *pod.resources("limits", debug=context.debug).as_tuple(),
            )


@table(schema="kubernetes", name="jobs", resource="jobs")
class JobsTable:

    @property
    def schema(self):
        return """
            name TEXT,
            uid TEXT,
            namespace TEXT,
            status TEXT,
            cpu_req REAL,
            gpu_req REAL,
            mem_req INTEGER,
            cpu_lim REAL,
            gpu_lim REAL,
            mem_lim INTEGER
        """

    def make_rows(self, context) -> list[tuple[dict, tuple]]:
        for item in context.data["items"]:
            job = JobHelper(item)
            yield item, (
                job.name,
                job.metadata.get("uid"),
                job.namespace,
                job.status,
                *job.resources("requests", debug=context.debug).as_tuple(),
                *job.resources("limits", debug=context.debug).as_tuple(),
            )


class LabelsTable:
    """Base class for all built-in label tables; subclasses need only define UID_FIELD."""

    @property
    def schema(self):
        return f"""
            {self.UID_FIELD} TEXT,
            key TEXT,
            value TEXT
        """

    def make_rows(self, context) -> list[tuple[dict, tuple]]:
        for item in context.data["items"]:
            thing = ItemHelper(item)
            for key, value in thing.labels.items():
                yield item, (thing.metadata.get("uid"), key, value)


@table(schema="kubernetes", name="node_labels", resource="nodes")
class NodeLabelsTable(LabelsTable):
    UID_FIELD = "node_uid"


@table(schema="kubernetes", name="pod_labels", resource="pods")
class PodLabelsTable(LabelsTable):
    UID_FIELD = "pod_uid"


@table(schema="kubernetes", name="job_labels", resource="jobs")
class JobLabelsTable(LabelsTable):
    UID_FIELD = "job_uid"


"""
Tests for command-line options.
"""

import sqlite3

import pytest

from kugl.main import main1
from kugl.util import KuglError


def test_enforce_one_cache_option(test_home):
    with pytest.raises(KuglError, match="Cannot use both -c/--cache and -u/--update"):
        main1(["-c", "-u", "select 1"])


def test_enforce_one_namespace_option(test_home):
    with pytest.raises(KuglError, match="Cannot use both -a/--all-namespaces and -n/--namespace"):
        main1(["-a", "-n", "x", "select * from pods"])


def test_no_such_table(test_home):
    with pytest.raises(sqlite3.OperationalError, match="no such table: foo"):
        main1(["select * from foo"])


def test_unknown_shortcut(test_home):
    with pytest.raises(KuglError, match="No shortcut named 'foo'"):
        main1(["foo"])


def test_shortcut_with_invalid_option(test_home, capsys):
    test_home.joinpath("init.yaml").write_text("""
        shortcuts:
          foo:
          - --badoption
          - "select * from pods"
    """)
    with pytest.raises(SystemExit):
        main1(["-a", "foo"])
    assert "unrecognized arguments: --badoption" in capsys.readouterr().err


def test_unknown_option(test_home, capsys):
    with pytest.raises(SystemExit):
        main1(["--badoption", "select 1"])
    assert "unrecognized arguments: --badoption" in capsys.readouterr().err


def test_enforce_one_cache_option_via_shortcut(test_home, capsys):
    test_home.joinpath("init.yaml").write_text("""
        shortcuts:
          foo:
          - -u
          - "select 1"
    """)
    with pytest.raises(KuglError, match="Cannot use both -c/--cache and -u/--update"):
        main1(["-c", "foo"])


def test_simple_shortcut(test_home, capsys):
    test_home.joinpath("init.yaml").write_text("""
        shortcuts:
          foo: ["select 1, 2"]
    """)
    main1(["foo"])
    out, _ = capsys.readouterr()
    assert out == "  1    2\n" * 2



"""
Pydantic models for configuration files.
"""
import json
import re
from typing import Literal, Optional, Tuple, Callable, Union

import jmespath
from pydantic import BaseModel, ConfigDict, ValidationError
from pydantic.functional_validators import model_validator

from kugl.util import Age, parse_utc, parse_size, KPath, ConfigPath, parse_age, parse_cpu, fail

PARENTED_PATH = re.compile(r"^(\^*)(.*)")


class Settings(BaseModel):
    """Holds the settings: entry from a user config file."""
    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True)
    cache_timeout: Age = Age(120)
    reckless: bool = False


class UserInit(BaseModel):
    """The root model for init.yaml; holds the entire file content."""
    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True)
    settings: Optional[Settings] = Settings()
    shortcuts: dict[str, list[str]] = {}


class ColumnDef(BaseModel):
    """Holds one entry from a columns: list in a user config file."""
    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True)
    name: str
    type: Literal["text", "integer", "real", "date", "age", "size", "cpu"] = "text"
    path: Optional[str] = None
    label: Optional[Union[str, list[str]]] = None
    # Function to extract a column value from an object.
    _extract: Callable[[object], object]
    # Function to convert the extracted value to the SQL type
    _convert: type
    # Parsed value of self.path
    _finder: jmespath.parser.Parser
    # Number of ^ in self.path
    _parents: int
    # SQL type for this column
    _sqltype: str

    @model_validator(mode="after")
    @classmethod
    def gen_extractor(cls, config: 'ColumnDef') -> 'ColumnDef':
        """
        Generate the extract function for a column definition; given an object, it will
        return a column value of the appropriate type.
        """
        if config.path and config.label:
            raise ValueError("cannot specify both path and label")
        elif config.path:
            m = PARENTED_PATH.match(config.path)
            config._parents = len(m.group(1))
            try:
                config._finder = jmespath.compile(m.group(2))
            except jmespath.exceptions.ParseError as e:
                raise ValueError(f"invalid JMESPath expression {m.group(2)} in column {config.name}") from e
            config._extract = config._extract_jmespath
        elif config.label:
            if not isinstance(config.label, list):
                config.label = [config.label]
            config._extract = config._extract_label
        else:
            raise ValueError("must specify either path or label")
        config._sqltype = KUGL_TYPE_TO_SQL_TYPE[config.type]
        config._convert = KUGL_TYPE_CONVERTERS[config.type]
        return config

    def extract(self, obj: object, context) -> object:
        """Extract the column value from an object and convert to the correct type."""
        if obj is None:
            if context.debug:
                print(f"No object provided to extractor {self}")
            return None
        if context.debug:
            print(f"Extract {self} from {self._abbreviate(obj)}")
        value = self._extract(obj, context)
        if context.debug:
            print(f"Extracted {value}")
        return None if value is None else self._convert(value)

    def _extract_jmespath(self, obj: object, context) -> object:
        """Extract a value from an object using a JMESPath finder."""
        if self._parents > 0:
            obj = context.get_parent(obj, self._parents)
        if obj is None:
            fail(f"Missing parent or too many ^ while evaluating {self.path}")
        return self._finder.search(obj)

    def _extract_label(self, obj: object, context) -> object:
        """Extract a value from an object using a label."""
        obj = context.get_root(obj)
        if available := obj.get("metadata", {}).get("labels", {}):
            for label in self.label:
                if (value := available.get(label)) is not None:
                    return value

    def __str__(self):
        if self.path:
            return f"{self.name} path={self.path}"
        return f"{self.name} label={','.join(self.label)}"

    def _abbreviate(self, obj):
        text = json.dumps(obj)
        if len(text) > 100:
            return text[:100] + "..."
        return text


KUGL_TYPE_CONVERTERS = {
    "integer": int,
    "real" : float,
    "text": str,
    "date": parse_utc,
    "age": parse_age,
    "size": parse_size,
    "cpu": parse_cpu,
}

KUGL_TYPE_TO_SQL_TYPE = {
    "integer": "integer",
    "real": "real",
    "text": "text",
    "date": "integer",
    "age": "integer",
    "size": "integer",
    "cpu": "real",
}


class ExtendTable(BaseModel):
    """Holds the extend: section from a user config file."""
    model_config = ConfigDict(extra="forbid")
    table: str
    columns: list[ColumnDef] = []


class ResourceDef(BaseModel):
    """Holds one entry from the resources: list in a user config file."""
    name: str
    # FIXME: Don't conflate all resource attributes in one class
    namespaced: bool = True
    cacheable: bool = True
    file: Optional[str] = None
    exec: Optional[Union[str, list[str]]] = None

    @model_validator(mode="after")
    @classmethod
    def validate(cls, config: 'ResourceDef') -> 'ResourceDef':
        if config.file and config.exec:
            raise ValueError("Resource cannot specify both file and exec")
        if config.file:
            config.cacheable = False
        return config

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return self.name == other.name


class CreateTable(ExtendTable):
    """Holds the create: section from a user config file."""
    resource: str
    row_source: Optional[list[str]] = None


class UserConfig(BaseModel):
    """The root model for a user config file; holds the complete file content."""
    model_config = ConfigDict(extra="forbid")
    resources: list[ResourceDef] = []
    extend: list[ExtendTable] = []
    create: list[CreateTable] = []


# FIXME use typevars
def parse_model(model_class, root: dict) -> Tuple[object, list[str]]:
    """Parse a dict into a model instance (typically a UserConfig).

    :return: A tuple of (parsed object, list of errors).  On success, the error list is None.
        On failure, the parsed object is None.
    """
    try:
        return model_class.model_validate(root), None
    except ValidationError as e:
        error_location = lambda err: '.'.join(str(x) for x in err['loc'])
        return None, [f"{error_location(err)}: {err['msg']}" for err in e.errors()]

# FIXME use typevars
def parse_file(model_class, path: ConfigPath) -> Tuple[object, list[str]]:
    """Parse a configuration file into a model instance, handling edge cases.

    :return: Same as parse_model."""
    if not path.exists():
        return model_class(), None
    if path.is_world_writeable():
        return None, [f"{path} is world writeable, refusing to run"]
    return parse_model(model_class, path.parse_yaml() or {})




"""
Process Kugl queries.
If you're looking for Kugl's "brain", you've found it.
See also tables.py
"""

from concurrent.futures import ThreadPoolExecutor
import json
from pathlib import Path
import re
import sys
from typing import Tuple, Set, Optional, Literal

import yaml
from pydantic import BaseModel, ConfigDict, Field
from tabulate import tabulate

from .config import ResourceDef, Settings
from .registry import Schema
from kugl.util import fail, SqliteDb, to_size, to_utc, kugl_home, clock, debugging, to_age, run, Age, KPath

# Needed to locate the built-in table builders by class name.
import kugl.builtins.kubernetes

# Cache behaviors
# TODO consider an enum

ALWAYS_UPDATE, CHECK, NEVER_UPDATE = 1, 2, 3
CacheFlag = Literal[ALWAYS_UPDATE, CHECK, NEVER_UPDATE]


class Query(BaseModel):
    """A SQL query + query-related behaviors"""
    sql: str
    # TODO: move this elsewhere, it's K8S-specific
    namespace: str = "default"
    cache_flag: CacheFlag = ALWAYS_UPDATE

    @property
    def table_refs(self) -> Set["TableRef"]:
        # Determine which tables are needed for the query by looking for symmbols that follow
        # FROM and JOIN.  Some of these may be CTEs, so don't assume they're all availabie in
        # Kubernetes, just pick out the ones we know about and let SQLite take care of
        # "unknown table" errors.
        # FIXME: use sqlparse package
        sql = self.sql.replace("\n", " ")
        refs = set(re.findall(r"(?<=from|join)\s+([.\w]+)", sql, re.IGNORECASE))
        return {TableRef.parse(ref) for ref in refs}

    @property
    def sql_schemaless(self) -> str:
        """Return the SQL query with schema hints removed."""
        sql = self.sql.replace("\n", " ")
        return re.sub(r"((from|join)\s+)[^.\s]+\.", r"\1", sql, flags=re.IGNORECASE)


class TableRef(BaseModel):
    """A reference to a table in a query."""
    model_config = ConfigDict(frozen=True)
    schema_name: str = Field(..., alias="schema")  # e.g. "kubernetes"
    name: str  # e.g. "pods"

    @classmethod
    def parse(cls, ref: str):
        """Parse a table reference of the form "pods" or "kubernetes.pods".
        SQLite doesn't actually support schemas, so the schema is just a hint.
        We replace the dot with an underscore to make it a valid table name."""
        parts = ref.split(".")
        if len(parts) == 1:
            return cls(schema="kubernetes", name=parts[0])
        if len(parts) == 2:
            if parts[0] == "k8s":
                parts[0] = "kubernetes"
            return cls(schema=parts[0], name=parts[1])
        fail(f"Invalid table reference: {ref}")


class Engine:

    def __init__(self, schema: Schema, settings: Settings, context_name: str):
        """
        :param config: the parsed user settings file
        :param context_name: the Kubernetes context to use, e.g. "minikube", "research-cluster"
        """
        self.schema = schema
        self.settings = settings
        self.context_name = context_name
        self.cache = DataCache(kugl_home() / "cache" / self.context_name, self.settings.cache_timeout)
        # Maps resource name e.g. "pods" to the response from "kubectl get pods -o json"
        self.data = {}
        self.db = SqliteDb()
        add_custom_functions(self.db.conn)

    def query_and_format(self, query: Query) -> str:
        """Execute a Kugl query and format the rsults for stdout."""
        rows, headers = self.query(query)
        return tabulate(rows, tablefmt="plain", floatfmt=".1f", headers=headers)

    def query(self, query: Query) -> Tuple[list[Tuple], list[str]]:
        """Execute a Kugl query but don't format the results.
        :return: a tuple of (rows, column names)
        """

        # Reconcile tables created / extended in the config file with tables defined in code, and
        # generate the table builders.
        tables = {}
        for name in {t.name for t in query.table_refs}:
            # Some of the named tables may be CTEs, so it's not an error if we can't create
            # them.  If actually missing when we reach the query, let SQLite issue the error.
            if (table := self.schema.table_builder(name)) is not None:
                tables[name] = table

        # Identify what to fetch vs what's stale or expired.
        resources_used = self.schema.resources_used(tables.values())
        refreshable, max_staleness = self.cache.advise_refresh(query.namespace, resources_used, query.cache_flag)
        if not self.settings.reckless and max_staleness is not None:
            print(f"(Data may be up to {max_staleness} seconds old.)", file=sys.stderr)
            clock.CLOCK.sleep(0.5)

        # Retrieve resource data in parallel.  If fetching from Kubernetes, update the cache;
        # otherwise just read from the cache.
        def fetch(res: ResourceDef):
            try:
                if res in refreshable:
                    self.data[res.name] = self._get_objects(res)
                    self.cache.dump(query.namespace, res.name, self.data[res.name])
                else:
                    self.data[res.name] = self.cache.load(query.namespace, res.name)
            except Exception as e:
                fail(f"Failed to get {res.name} objects: {e}")
        with ThreadPoolExecutor(max_workers=8) as pool:
            for _ in pool.map(fetch, resources_used):
                pass

        # Create tables in SQLite
        for table in tables.values():
            table.build(self.db, self.data[table.resource])

        column_names = []
        rows = self.db.query(query.sql, names=column_names)
        # %g is susceptible to outputting scientific notation, which we don't want.
        # but %f always outputs trailing zeros, which we also don't want.
        # So turn every value x in each row into an int if x == float(int(x))
        truncate = lambda x: int(x) if isinstance(x, float) and x == float(int(x)) else x
        rows = [[truncate(x) for x in row] for row in rows]
        return rows, column_names

    def _get_objects(self, resource: ResourceDef) -> dict:
        """Called for a cache miss, go actually fetch resources.
        Handle built-in resources here, and dispatch to schema implementation for those
        that aren't.
        """
        def parse(text):
            if not text:
                return {}
            if text[0] in "{[":
                return json.loads(text)
            return yaml.safe_load(text)

        if resource.file is not None:
            if resource.file == "stdin":
                return parse(sys.stdin.read())
            try:
                return parse(Path(resource.file).read_text())
            except OSError as e:
                fail(f"Failed to read {resource.file}", e)

        if resource.exec is not None:
            _, out, _ = run(resource.exec)
            return parse(out)

        return self.schema.impl.get_objects(resource.name, resource.namespaced)


class DataCache:
    """Manage the cached JSON data from Kubectl.
    This is a separate class for ease of unit testing.
    """

    def __init__(self, dir: KPath, timeout: Age):
        """
        :param dir: root of the cache folder tree; paths are of the form
            <kubernetes context>/<namespace>.<resource kind>.json
        :param timeout: age at which cached data is considered stale
        """
        self.dir = dir
        dir.mkdir(parents=True, exist_ok=True)
        self.timeout = timeout

    def advise_refresh(self, namespace: str, resources: Set[ResourceDef], flag: CacheFlag) -> Tuple[Set[str], int]:
        """Determine which resources to use from cache or to refresh.

        :param namespace: the Kubernetes namespace to query, or ALL_NAMESPACE
        :param resources: the resource types to consider
        :param flag: the user-specified cache behavior
        :return: a tuple of (refreshable, max_age) where refreshable is the set of resources types
            to update, and max_age is the maximum age of the resources that won't be updated.
        """
        if flag == ALWAYS_UPDATE:
            # Refresh everything and don't issue a "stale data" warning
            return resources, None
        # Find what's expired or missing
        cacheable = {r for r in resources if r.cacheable}
        non_cacheable = {r for r in resources if not r.cacheable}
        cache_ages = {r: self.age(self.cache_path(namespace, r.name)) for r in cacheable}
        expired = {r for r, age in cache_ages.items() if age is not None and age >= self.timeout.value}
        missing = {r for r, age in cache_ages.items() if age is None}
        # Always refresh what's missing or non-cacheable, and possibly also what's expired
        # Stale data warning for everything else
        refreshable = missing if flag == NEVER_UPDATE else expired | missing
        max_age = max((cache_ages[r] for r in (cacheable - refreshable)), default=None)
        refreshable.update(non_cacheable)
        if debugging("cache"):
            print("Requested", [r.name for r in resources])
            print("Cacheable", [r.name for r in cacheable])
            print("Non-cacheable", [r.name for r in non_cacheable])
            print("Ages", " ".join(f"{r.name}={age}" for r, age in cache_ages.items()))
            print("Expired", [r.name for r in expired])
            print("Missing", [r.name for r in missing])
            print("Refreshable", [r.name for r in refreshable])
        return refreshable, max_age

    def cache_path(self, namespace: str, kind: str) -> Path:
        return self.dir / f"{namespace}.{kind}.json"

    def dump(self, namespace: str, kind: str, data: dict):
        self.cache_path(namespace, kind).write_text(json.dumps(data))

    def load(self, namespace: str, kind: str) -> dict:
        return json.loads(self.cache_path(namespace, kind).read_text())

    def age(self, path: Path) -> Optional[int]:
        """The age of a file in seconds, relative to the current time, or None if it doesn't exist."""
        if not path.exists():
            if debugging("cache"):
                print("Missing cache file", path)
            return None
        age_secs = int(clock.CLOCK.now() - path.stat().st_mtime)
        if debugging("cache"):
            print(f"Found cache file (age = {to_age(age_secs)})", path)
        return age_secs


def add_custom_functions(db):
    db.create_function("to_size", 1, lambda x: to_size(x, iec=True))
    # This must be a lambda because the clock is patched in unit tests
    db.create_function("now", 0, lambda: clock.CLOCK.now())
    db.create_function("to_age", 1, to_age)
    db.create_function("to_utc", 1, to_utc)





"""
Wrappers to make JSON returned by kubectl easier to work with.
"""

from abc import abstractmethod
from dataclasses import dataclass
from typing import Optional

import funcy as fn

from kugl.util import parse_size, parse_cpu

# What container name is considered the "main" container, if present
MAIN_CONTAINERS = ["main", "notebook", "app"]


@dataclass
class Limits:
    """
    A class to hold CPU, GPU and memory resources. This is called "Limits" although it's used for both requests
    and limits, so as not to confuse "resources" with Kubernetes resources in general.
    """
    cpu: Optional[float]
    gpu: Optional[float]
    mem: Optional[int]

    def __add__(self, other):
        if self.cpu is None and other.cpu is None:
            cpu = None
        else:
            cpu = (self.cpu or 0) + (other.cpu or 0)
        if self.gpu is None and other.gpu is None:
            gpu = None
        else:
            gpu = (self.gpu or 0) + (other.gpu or 0)
        if self.mem is None and other.mem is None:
            mem = None
        else:
            mem = (self.mem or 0) + (other.mem or 0)
        return Limits(cpu, gpu, mem)

    def __radd__(self, other):
        """Needed to support sum() -- handles 0 as a starting value"""
        return self if other == 0 else self.__add__(other)

    def as_tuple(self):
        return (self.cpu, self.gpu, self.mem)

    @classmethod
    def extract(cls, obj, debug=False):
        """Extract a Limits object from a dictionary, or return an empty one if the dictionary is None.

        :param obj: A dictionary with keys "cpu", "nvidia.com/gpu" and "memory" """
        if obj is None:
            if debug:
                print("No object provided to Limits extractor")
            return Limits(None, None, None)
        if debug:
            print("Extracting limits from", obj)
        cpu = parse_cpu(obj.get("cpu"))
        gpu = parse_cpu(obj.get("nvidia.com/gpu"))
        mem = parse_size(obj.get("memory"))
        result = Limits(cpu, gpu, mem)
        if debug:
            print("Extracted", result)
        return result


class ItemHelper:
    """Some common code for wrappers on JSON for pods, nodes et cetera."""

    def __init__(self, obj):
        self.obj = obj
        self.metadata = self.obj.get("metadata", {})
        self.labels = self.metadata.get("labels", {})

    def __getitem__(self, key):
        """Return a key from the object; no default, will error if not present"""
        return self.obj[key]

    @property
    def name(self):
        """Return the name of the object from the metadata, or none if unavailable."""
        return self.metadata.get("name") or self.obj.get("name")

    @property
    def namespace(self):
        """Return the name of the object from the metadata, or none if unavailable."""
        return self.metadata.get("namespace")

    def label(self, name):
        """Return one of the labels from the object, or None if it doesn't have that label."""
        return self.labels.get(name)


class Containerized:

    @abstractmethod
    def containers(self):
        raise NotImplementedError()

    def resources(self, tag, debug=False):
        return sum(Limits.extract(c.get("resources", {}).get(tag), debug) for c in self.containers)


class PodHelper(ItemHelper, Containerized):

    @property
    def command(self):
        return " ".join((self.main or {}).get("command", []))

    @property
    def is_daemon(self):
        return any(ref.get("kind") == "DaemonSet" for ref in self.metadata.get("ownerReferences", []))

    @property
    def containers(self):
        """Return the containers in the pod, if any, else an empty list."""
        return self["spec"].get("containers", [])

    @property
    def main(self):
        """Return the main container in the pod, if any, defined as the first container with a name
        in MAIN_CONTAINERS.  If there are none of those, return the first one.
        """
        if not self.containers:
            return None
        main = fn.first(fn.filter(lambda c: c["name"] in MAIN_CONTAINERS, self.containers))
        return main or self.containers[0]


class JobHelper(ItemHelper, Containerized):

    @property
    def status(self):
        status = self.obj.get("status", {})
        if len(status) == 0:
            return "Unknown"
        # Per
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1JobStatus.md
        # and https://kubernetes.io/docs/concepts/workloads/controllers/job/
        for c in status.get("conditions", []):
            if c["status"] == "True":
                if c["type"] == "Failed":
                    # TODO use a separate column
                    return c.get("reason") or "Failed"
                if c["type"] == "Suspended":
                    return "Suspended"
                if c["type"] == "Complete":
                    return "Complete"
            if c["type"] == "FailureTarget":
                return "Failed"
            if c["type"] == "SuccessCriteriaMet":
                return "Complete"
        if status.get("active", 0) > 0:
            return "Running"
        return "Unknown"

    @property
    def containers(self):
        """Return the containers in the job, if any, else an empty list."""
        return self["spec"]["template"]["spec"].get("containers", [])




"""
Tests for the pods table.
"""

import pytest

from kugl.builtins.helpers import PodHelper
from kugl.util import UNIT_TEST_TIMEBASE

from .testing import make_pod, kubectl_response, Container, CGM, assert_query, make_job


def test_missing_metadata():
    """Verify pod.label() does not fail if metadata is missing"""
    pod = PodHelper(make_pod("noname", no_metadata=True))
    assert pod.label("foo") is None


def test_name():
    """Verify different locations for the pod name, including no name at all."""
    pod = PodHelper(make_pod("mypod-1"))
    assert pod.name == "mypod-1"
    pod = PodHelper(make_pod("mypod-2", name_at_root=True))
    assert pod.name == "mypod-2"
    pod = PodHelper(make_pod("mypod-3", no_name=True))
    assert pod.name is None


def test_no_main_container():
    # Not sure when this would apply; it's here for coverage
    pod = make_pod("pod-1", containers=[])
    assert PodHelper(pod).main is None


def test_by_cpu(test_home):
    kubectl_response("pods", {
        "items": [
            make_pod("pod-1"),
            make_pod("pod-2"),
            make_pod("pod-3", containers=[Container(requests=CGM(cpu=2, mem="10M"))]),
            make_pod("pod-4", containers=[Container(requests=CGM(cpu="2000m", mem="10M"))]),
            # should get dropped because no status available
            make_pod("pod-5", containers=[Container(requests=CGM(cpu=2, mem="10M"))]),
        ]
    })
    kubectl_response("pod_statuses", """
        NAME   STATUS
        pod-1  Init:1
        pod-2  Init:2
        pod-3  Init:3
        pod-4  Init:4
    """)
    assert_query("SELECT name, status FROM pods WHERE cpu_req > 1 ORDER BY name", """
        name    status
        pod-3   Init:3
        pod-4   Init:4
    """)


def test_other_pod_fields(test_home):
    kubectl_response("pods", {
        "items": [
            make_pod("pod-1", namespace="xyz", is_daemon=True, phase="Pending"),
            make_pod("pod-3", node_name="joe", creation_ts=UNIT_TEST_TIMEBASE + 60),
            make_pod("pod-4", containers=[Container(command=["echo", "bye"])]),
        ]
    })
    kubectl_response("pod_statuses", """
        NAMESPACE  NAME   STATUS
        xyz        pod-1  Pending
        default    pod-2  Running
        default    pod-3  Terminating
        default    pod-4  Terminating
    """)
    assert_query("""
        SELECT namespace, phase, uid, is_daemon, node_name, command, to_utc(creation_ts) AS created
        FROM pods ORDER BY name
    """, """
        namespace    phase    uid          is_daemon  node_name    command     created
        xyz          Pending  uid-pod-1            1  worker5      echo hello  2024-12-10T02:49:02Z
        default      Running  uid-pod-3            0  joe          echo hello  2024-12-10T02:50:02Z
        default      Running  uid-pod-4            0  worker5      echo bye    2024-12-10T02:49:02Z
    """, all_ns=True)


@pytest.mark.parametrize("containers,expected", [
    # Typical pod - cpu/mem requests/limits, no GPU
    ([Container(requests=CGM(cpu=1, mem="1Mi"), limits=CGM(cpu=1, mem="1Mi"))],
     [ [1, 1, 1<<20, 1<<20, None, None] ]),
    # Same thing but remove limits
    ([Container(requests=CGM(cpu=1, mem="1Mi"), limits=CGM())],
     [ [1, None, 1<<20, None, None, None] ]),
    # Repeat first example but with two containers; should sum the GPU Nones
    ([Container(requests=CGM(cpu=1, mem="1Mi"), limits=CGM(cpu=1, mem="1Mi")),
      Container(requests=CGM(cpu=1, mem="1Mi"), limits=CGM(cpu=1, mem="1Mi"))],
     [ [2, 2, 2<<20, 2<<20, None, None] ]),
    # Add a GPU request
    ([Container(requests=CGM(cpu=1, mem="1Mi", gpu=3), limits=CGM(cpu=1, mem="1Mi", gpu=3))],
     [ [1, 1, 1<<20, 1<<20, 3, 3] ]),
    # Mixing set and unset limits should result in Nones
    ([Container(requests=CGM(cpu=1, mem="1Mi"), limits=CGM(cpu=1, mem="1Mi")),
      Container(requests=CGM(cpu=1, mem="1Mi"), limits=CGM(cpu=1, mem="1Mi")),
      Container(requests=CGM(cpu=1, mem="1Mi"), limits=CGM())],
     [ [3, 2, 3<<20, 2<<20, None, None] ]),
])
def test_resource_summing(test_home, containers, expected):
    pod = make_pod("pod-1", containers=containers)
    kubectl_response("pods", {"items": [pod]})
    kubectl_response("pod_statuses", "NAME    STATUS\npod-1  Running")
    assert_query("SELECT cpu_req, cpu_lim, mem_req, mem_lim, gpu_req, gpu_lim FROM pods", expected)
    # Try the same thing but with a job
    job = make_job("job-1", pod=pod)
    kubectl_response("jobs", {"items": [job]})
    assert_query("SELECT cpu_req, cpu_lim, mem_req, mem_lim, gpu_req, gpu_lim FROM jobs", expected)


def test_pod_labels(test_home):
    kubectl_response("pods", {
        "items": [
            make_pod("pod-1", labels=dict(foo="bar")),
            make_pod("pod-2", labels=dict(a="b", c="d", e="f")),
            make_pod("pod-3", labels=dict()),
            make_pod("pod-4", labels=dict(one="two", three="four")),
        ]
    })
    kubectl_response("pod_statuses", """
        NAME   STATUS
        pod-1  Init:1
        pod-2  Init:2
        pod-3  Init:3
        pod-4  Init:4
    """)
    assert_query("SELECT pod_uid, key, value FROM pod_labels ORDER BY 2, 1", """
        pod_uid    key    value
        uid-pod-2  a      b
        uid-pod-2  c      d
        uid-pod-2  e      f
        uid-pod-1  foo    bar
        uid-pod-4  one    two
        uid-pod-4  three  four
    """)



"""
Imports usable by user-defined tables in Python (once we have those.)
"""

from kugl.impl.registry import Registry

from kugl.util import (
    fail,
    parse_age,
    parse_utc,
    to_age,
    to_utc,
)


def schema(name: str):
    def wrap(cls):
        Registry.get().add_schema(name, cls)
        return cls
    return wrap


def table(**kwargs):
    def wrap(cls):
        Registry.get().add_table(cls, **kwargs)
        return cls
    return wrap




from pathlib import Path
from setuptools import setup, find_packages
from textwrap import dedent

setup(
    name="kugl",
    version="0.3.2",
    description="Explore Kubernetes resources using SQLite",
    long_description=dedent("""
    PyPI readme goes here.
    
    For documentation please visit the [GitHub repository](https://github.com/jonross/kugl).
    """),
    long_description_content_type="text/markdown",
    author="Jon Ross",
    author_email="kugl.devel@gmail.com",
    url="https://github.com/jonross/kugl",
    packages=find_packages(),
    include_package_data=True,
    install_requires=(Path(__file__).parent / "reqs_public.txt").read_text().strip().splitlines(),
    entry_points={
        "console_scripts": [
            "kugl = kugl.main:main",
        ],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",  # Minimum Python version
)




"""
Registry of resources and tables, independent of configuration file format.
This is Kugl's global state outside the SQLite database.
"""
from argparse import ArgumentParser
from typing import Type, Iterable

from pydantic import BaseModel

from kugl.impl.config import UserConfig, parse_file, CreateTable, ExtendTable, ResourceDef
from kugl.impl.tables import TableFromCode, TableFromConfig, TableDef, Table
from kugl.util import fail, dprint, ConfigPath, kugl_home

_REGISTRY = None


class Registry:
    """All known tables and resources.
    There is one instance of this in any Kugl process."""

    def __init__(self):
        self.schemas: dict[str, Schema] = {}

    @staticmethod
    def get():
        global _REGISTRY
        if _REGISTRY is None:
            _REGISTRY = Registry()
        return _REGISTRY

    def add_schema(self, name: str, cls: Type):
        """Register a class to implement a schema; this is called by the @schema decorator."""
        dprint("registry", f"Add schema {name} {cls}")
        self.schemas[name] = Schema(name=name, impl=cls())

    def get_schema(self, name: str) -> "Schema":
        if name not in self.schemas:
            self.add_schema(name, GenericSchemaImpl)
        return self.schemas[name].augment()

    def add_table(self, cls, **kwargs):
        """Register a class to define a table in Python; this is called by the @table decorator."""
        dprint("registry", f"Add table {kwargs}")
        t = TableDef(cls=cls, **kwargs)
        if t.schema_name not in self.schemas:
            fail(f"Must create schema {t.schema_name} before table {t.schema_name}.{t.name}")
        self.schemas[t.schema_name].builtin[t.name] = t


class Schema(BaseModel):
    """Collection of tables and resource definitions.

    Capture a schema definition from the @schema decorator, example:
        @schema("kubernetes")
    Or, capture a schema definition in a user config file.
    Or both.
    """
    name: str
    impl: object # FIXME use type vars
    builtin: dict[str, TableDef] = {}
    _create: dict[str, CreateTable] = {}
    _extend: dict[str, ExtendTable] = {}
    _resources: dict[str, ResourceDef] = {}

    def augment(self):
        """Apply the built-in and user configuration files for the schema, if present."""
        def _apply(path: ConfigPath):
            if path.exists():
                config, errors = parse_file(UserConfig, path)
                if errors:
                    fail("\n".join(errors))
                self._create.update({c.table: c for c in config.create})
                self._extend.update({e.table: e for e in config.extend})
                self._resources.update({r.name: r for r in config.resources})

        # Reset the non-builtin tables, since these can change during unit tests.
        for mapping in [self._create, self._extend, self._resources]:
            mapping.clear()

        # Apply builtin config and user config
        _apply(ConfigPath(__file__).parent.parent / "builtins" / f"{self.name}.yaml")
        _apply(ConfigPath(kugl_home() / f"{self.name}.yaml"))

        # Verify user-defined tables have the needed resources
        for table in self._create.values():
            if table.resource not in self._resources:
                fail(f"Table '{table.table}' needs unknown resource '{table.resource}'")

        return self

    def table_builder(self, name):
        """Return the Table builder subclass (see tables.py) for a table name."""
        builtin = self.builtin.get(name)
        creator = self._create.get(name)
        extender = self._extend.get(name)
        if builtin and creator:
            fail(f"Pre-defined table {name} can't be created from config")
        if builtin:
            return TableFromCode(builtin, extender)
        if creator:
            return TableFromConfig(name, creator, extender)

    def resources_used(self, tables: Iterable[Table]) -> set[ResourceDef]:
        """Return the ResourceDefs used by the listed tables."""
        return {self._resources[t.resource] for t in tables}


class GenericSchemaImpl:
    """get_schema auto-generates one of these when an undefined schema is referenced."""

    def add_cli_options(self, ap: ArgumentParser):
        # FIXME, artifact of assuming kubernetes
        self.ns = "default"
        pass

    def handle_cli_options(self, args):
        pass


"""
Our entire interface to SQLite features is here.
"""

import collections as co
import sqlite3

from kugl.util import dprint


class SqliteDb:

    def __init__(self, target=None):
        self.target = target
        self.conn = sqlite3.connect(":memory:", check_same_thread=False) if target is None else None

    def query(self, sql, **kwargs):
        """
        Query boilerplate reducer.
        :param sql str: SQL query
        :param data list: Optional query args
        :param one_row bool: If True, use cursor.fetchone() instead of fetchall()
        :param named bool: If True, rows are namedtuples
        :param names list: If an array, append column names to it
        """
        dprint("sqlite", f"Query: {sql}")
        if self.conn:
            return self._query(self.conn, sql, **kwargs)
        else:
            with sqlite3.connect(self.target) as conn:
                return self._query(conn, sql, **kwargs)

    def _query(self, conn, sql, data=None, named=False, names=None, one_row=False):
        cur = conn.cursor()
        res = cur.execute(sql, data or [])
        if names is not None:
            names.extend(col[0] for col in cur.description)
        if named:
            Row = co.namedtuple("Row", [col[0] for col in cur.description])
            if one_row:
                row = cur.fetchone()
                return row and Row(*row)
            else:
                rows = cur.fetchall()
                return [Row(*row) for row in rows]
        else:
            if one_row:
                return cur.fetchone()
            else:
                return cur.fetchall()

    def execute(self, sql, data=None):
        """
        Non-query boilerplate reducer.
        :param sql str: SQL query
        :param data list: Optional update args
        """
        dprint("sqlite", f"Execute: {sql}")
        if self.conn:
            self._execute(self.conn, sql, data or [])
        else:
            with sqlite3.connect(self.target) as conn:
                self._execute(conn, sql, data or [])

    def _execute(self, conn, sql, data):
        if len(data) > 0 and any(isinstance(data[0], x) for x in [list, tuple]):
            conn.cursor().executemany(sql, data)
        else:
            conn.cursor().execute(sql, data)

